#!/usr/local/bin/python
# Rob Dennis - Lurchington Q. Delacroix
# Keasby - TV renamer
# Project Homepage: http://code.google.com/p/keasby/

#---Python Imports
from urlparse import urljoin,urlunsplit,urlsplit
from xml.dom import minidom
from string import Template
import random
import urllib2
import socket
import os
import sys
import glob
import zipfile
import anydbm
import shelve
import re
import logging
#---TVDB Imports
from tvdb_classes import tvDbMirror,tvDbSeries,tvDbEpisode

random.seed()
socket.setdefaulttimeout(8)

def sanitize_searchterm(searchterm):
    return re.sub("\W+"," ",searchterm).strip()

def getElements(dom,tagName,index=None):
    if dom is None:
        logging.warning("unable to get elements for empty DOM")
        return None
    
    if index==None:
            return dom.getElementsByTagName(tagName)
    else:
        try:
            return dom.getElementsByTagName(tagName)[index]
        except IndexError:
            return None
        
    
def getUrlRef(url):
    try:
        return urllib2.urlopen(url)
    except urllib2.URLError:
        logging.error("URLError caught, unable to open %s",url)
        raise
        

def addUrlQuery(splitUrl,query):
        return  urlunsplit(
                            (   splitUrl[0],
                                splitUrl[1],
                                splitUrl[2],
                                query,
                                splitUrl[4]
                                )
                          )

class TvDbApiHandler():
    def __init__(self,apiKey):
        self.domDict = {}
        self.init_errors()
        self.apiKey = apiKey
        self.baseUrl = r"http://www.thetvdb.com"
        self.cacheDir = os.path.join(os.path.dirname(sys.argv[0]),"cache")
        
        if not os.path.exists(self.cacheDir):
            os.makedirs(self.cacheDir)
        
        self.searchMap = anydbm.open(
                            os.path.join(self.cacheDir,"searchmap.cache"),
                                    "c")
        
        self.flush_series_list()
        
        if not os.path.exists(self.cacheDir):
            os.makedirs(self.cacheDir)
            
        try:
            if not os.path.exists(os.path.join(self.cacheDir,"languages.xml")):
                self.saveLanguageXml()
            self.set_language()
        except:
            logging.warning("unable to set language")
            logging.debug("defaulting to en")
            self.set_language(u"en")

#---Diagnostic
    def add_error(self,error_key,error_message):
        if error_key not in self.error_dict:
            logging.debug("adding error listing for %s",error_key)
            self.error_dict[error_key] = []
        
        logging.warning("error_key %s: %s",error_key,error_message)
        self.error_dict[error_key].append(error_message)
        
    def has_error(self,error_key):
        
        errors=self.error_dict.get(error_key)
        
##        if errors is not None:
##            logging.debug("%s has the following errors: %s",error_key,errors)
                                            
        
        api_error = self.error_dict.get(self.all_error_key)
        
        if api_error is None:
            api_error = []
        
        if error_key is self.all_error_key:
            return api_error
        
        else:
            current_error = self.error_dict.get(error_key)
        
            if current_error is None:
                current_error = []
                
            return current_error# + api_error
                
    
#---Initializations       
    def flush_series_list(self):
        logging.debug("creating series list")
        self.seriesList=shelve.open(os.path.join(self.cacheDir,"series_list.cache"),"c",writeback=True)
        
    def set_language(self,specified_language=None):
        if specified_language is None:
            self.language = u"en"
        else:
            self.language = specified_language

    def initMirrors(self):
        try:
            xmldoc = self.getXmlDomUrl(self.shapeUrl(self.baseUrl,
                                                            "mirrors.xml"))
        except urllib2.URLError:
            self.add_error(self.all_error_key,"unable to initialize mirrors")
            return None
            
        self.mirrors={"xml":[],"banner":[],"zip":[]}
        
        mirror_list = getElements(xmldoc,"Mirror")
        
        for mirror in mirror_list:
            tmp=tvDbMirror(mirror)
            typemask=int(tmp["typemask"])
            if typemask & 1 == 1:
                self.mirrors["xml"].append(tmp)
            if typemask & 2 == 2:
                self.mirrors["banner"].append(tmp)
            if typemask & 4 == 4:
                self.mirrors["zip"].append(tmp)

                
    def init_errors(self):
        self.error_dict = {}
        self.all_error_key = "TVDB_API"
        
        
#---Source Interfaces
    def get_title(self,showname,season_number,episode_number,
                  force_update=False):
                
        if isinstance(showname,unicode):
            string_showname = showname.encode("UTF-8")
        else:
            string_showname = showname
                
#Dumping on cached errors
        if self.has_error(self.all_error_key):
        
            logging.debug("previous error for %s: %s",
                         showname,
                         self.has_error(showname))
                        
            return None,None
        
        if self.has_error("%s-%s"%(showname,season_number)):
            logging.debug("previous error for %s-%s: %s",
                         showname,
                         season_number,
                         self.has_error(showname))
            
            return None,None
        
        if self.has_error(showname):
        
            logging.debug("previous error for %s: %s",
                         showname,
                         self.has_error(showname))
                        
            return None,None
        
        if self.has_error("%s-%s"%(showname,season_number)):
            logging.debug("previous error for %s-%s: %s",
                         showname,
                         season_number,
                         self.has_error(showname))
            
            return None,None
                                    
        
        logging.debug("showname %s",showname)

  
        seriesListing=self.seriesList.get(string_showname)
            
        init_season_order=False
        
        if force_update or seriesListing is None :
            logging.debug("showname %s not cached, querying",showname)
            try:
                seriesListing=self.get_full_query(showname)
            except Exception:
                self.add_error(showname,"error getting title for s:%s e:%s"%
                                              (season_number,episode_number))
            
                raise
                return None,None
                
            
            self.seriesList[string_showname]=seriesListing
            self.seriesList.sync()
            init_season_order=True
            
        if (not init_season_order 
            and not "SeasonNumber" in seriesListing.order):
            
            self.add_error(showname,"unable to correctly order SeasonNumber")
            logging.debug("initializing SeasonNumber order")
            seriesListing.order["SeasonNumber"] = {}
        
        if (force_update or init_season_order or not 
                    seriesListing.order["SeasonNumber"].get(season_number)):
                    
            if season_number > seriesListing.get_max_value("SeasonNumber"):
                logging.debug("season_number %s greater then max: %s",
                                season_number,
                                seriesListing.get_max_value("SeasonNumber"))
                return None,None
            
            logging.debug("season number %s not found, sorting",season_number)
            seriesListing.orderBy("SeasonNumber")
            self.seriesList[string_showname].order=seriesListing.order
            self.seriesList.sync()
        
        
        
        logging.debug("searching for episode_number %s",episode_number)
        
        try:
            season_list=seriesListing.order["SeasonNumber"][season_number]
        except KeyError:
            self.add_error("%s-%s"%(showname,season_number),
                           "error sorting by season %s"%season_number)
            return None,None
        
        if not season_list:
            self.add_error("%s-%s"%(showname,season_number),
                           "empty season %s list"%season_number)
            return None,None
            
        try:
            assumed_episode=season_list[episode_number-1]
        except IndexError:
            logging.debug("episode_number %s > length of %s",
                            episode_number,len(season_list))
            assumed_episode=season_list[-1]
                            
        if assumed_episode["EpisodeNumber"] == episode_number:                
        
            logging.debug("successfully assumed episode")
            return assumed_episode["EpisodeName"],seriesListing["SeriesName"]
                                                
        logging.debug("unable to assume episode, checking all season %s",
                                                            season_number)
        
        for episode in season_list:
            if episode["EpisodeNumber"] == episode_number:
                logging.debug("found title while checking all") 
                return episode["EpisodeName"],seriesListing["SeriesName"]
        
        if not force_update:
            logging.debug("title not found, forcing refresh")
            return self.get_title(showname,season_number,
                                    episode_number,force_update=True)
            
        
#---Queries
#Constructors for TVDB API Queries
    def queryURL_GetSeries(self,showname):
        baseURL=urlsplit(urljoin(self.randMirror("xml"),"api/GetSeries.php"))
        reshapeName=re.sub(r"\s+","%20",showname,re.U)
        return addUrlQuery(baseURL,
                    u"seriesname=%s&language=%s"%(reshapeName,self.language))

#Returning the appropriate information based on queries

    def qGetSeriesSearchDom(self,showname):

        queryURL=self.queryURL_GetSeries(showname)
        return self.getXmlDomUrl(queryURL)
    
        
    def qGetSeriesMapping(self,showname,cacheSearchXmlFlag=False):

        mappedID = self.getSearchMapping(showname)
        
        if not mappedID:
            shownameSearchDom = self.qGetSeriesSearchDom(showname)  
            chosenID=self.getSeriesId(shownameSearchDom,showname)
            logging.info("this is the chosen ID for %s: %s",
                             showname,
                             chosenID)
            
            if chosenID is None:
                logging.warning("showname %s isn't found, trying alternate",
                                showname)
                scrubbed_showname = sanitize_searchterm(showname)
                shownameSearchDom = self.qGetSeriesSearchDom(scrubbed_showname) 
                chosenID=self.getSeriesId(shownameSearchDom,showname)
                
                if chosenID is not None:
                    logging.info("this is the chosen ID for %s: %s",
                                     showname,
                                     chosenID)
                    logging.info("using: %s",scrubbed_showname)
                                
                else:
                    logging.warning("unable to find series ID using %s",
                                    scrubbed_showname)
                    return None
            
            
            
            self.addSearchMapping(showname,chosenID)
            
            if cacheSearchXmlFlag:
                out_filename=u"%s.search_results.xml"%showname
                queryURL=queryURL_GetSeries(showname)
                self.cacheFile(getUrlRef(queryURL),out_filename)
    
            return chosenID 
        
        else:
            return mappedID
        
    def qGetEpisodesList(self,showname,seriesDom=None):

        if seriesDom:
            episodeList=self.getEpisodeList(seriesDom)
        else:
            seriesAllZipFile=self.get_full_zip_dom_query(showname)
            episodeList=self.getEpisodeList(seriesAllZipFile)

        return episodeList
    
    def get_full_query(self,showname):

        try:
            seriesAllDom=self.get_full_zip_dom_query(showname)
        except:
            self.add_error(showname,"unable to get DOM from zip")
            raise

        series = getElements(seriesAllDom,"Series",0)
        episodeList=self.qGetEpisodesList(showname,seriesAllDom)
                
        return tvDbSeries(series,linkedInstances=episodeList)

    def get_full_zip_dom_query(self,showname):

        mapping=self.qGetSeriesMapping(showname)
        
        if self.has_error(showname):
            logging.warning("dumping from previous error:")
            logging.warning(self.has_error(showname))
            return None
        
        try:
            seriesAllZipFile = self.cacheSeriesAllZip(mapping,showname)
        except:
            self.add_error(showname,"unable to cache zip")
            raise
        
        return self.getXmlDomZip(seriesAllZipFile,"%s.xml"%self.language)
    
#---Parsing
#Parsing XML DOM for key tags

    def getSeriesId(self,xmldoc,showname):
        # print "assuming first result is desired at this time"
        
        series_list = getElements(xmldoc,"Series")
        
        seriesDomList=[tvDbSeries(series) for series in series_list]
        
        try:
            chosenID=seriesDomList[0]["seriesid"]
        except IndexError:
            self.add_error(showname,"series not found")
            return None
            
        return chosenID
        
    def getEpisodeList(self,xmldoc):
        episode_list = getElements(xmldoc,"Episode")
        return [tvDbEpisode(episode) for episode in episode_list]

#--- Data Persistence 
#       Search Terms
    def addSearchMapping(self,searchTerm,desiredOutcome):
        if isinstance(searchTerm,unicode):
            searchTerm = searchTerm.encode("UTF-8")
            
        self.searchMap[searchTerm]=str(desiredOutcome)
        
        
    def getSearchMapping(self,searchTerm):
        try:
            if isinstance(searchTerm,unicode):
                searchTerm = searchTerm.encode("UTF-8")
            mapping = self.searchMap[searchTerm]
            
        except KeyError:
            return None
        
        return mapping
        
#       Saving API downloaded items to disk
    def isCached(self,path):
        if self.cacheDir not in os.path.dirname(path):
            return os.path.exists(self.getCachePath(path))
        else:
            return os.path.exists(path)
    
    def getCachePath(self,path):
        if self.cacheDir not in os.path.dirname(path):
            return os.path.join(self.cacheDir,path)
        else:
            return path

    def cacheFile(self,fileObject,outputPath,newOutputFilename=None,overwriteFlag=True):
        if newOutputFilename:
            outputPath=os.path.join(os.path.dirname(outputPath),outputFilename)
            
        cachedPath = self.getCachePath(outputPath)

        if not os.path.exists(os.path.dirname(cachedPath)):
            os.makedirs(os.path.dirname(cachedPath))
            
        if overwriteFlag or not self.isCached(cachedPath):
            outFile = open(cachedPath,mode="wb")
            outFile.write(fileObject.read())
            outFile.close()
        
        return cachedPath
#       Convenience functions for caching common files
    def cacheSeriesAllZip(self,seriesID,showname):

        relativePath = os.path.join("series",
                                    str(seriesID),
                                    "all",
                                    "%s.zip"%self.language)
                                    
        url_path = r"series/%s/all/%s.zip"%(str(seriesID),self.language)

        if not os.path.exists(os.path.join(self.cacheDir,relativePath)):
            try:
                linkToZip=self.shapeUrl(self.randMirror("zip"),url_path)
            except:
                self.add_error(showname,"unable to get zip url from mirror")
                raise
                
            
            try:
                self.cacheFile(getUrlRef(linkToZip),relativePath)
            except urllib2.HTTPError:
                self.add_error(showname,"error opening URL:%s"%linkToZip)
                raise
                
        if self.has_error(showname):
            return None
            
        return zipfile.ZipFile(os.path.join(self.cacheDir,relativePath),"r")
    
    def saveLanguageXml(self):
        try:
            languageXmlUrl = getUrlRef(self.shapeUrl(
                                    self.randMirror("xml"),"languages.xml"))
        except AttributeError:
            self.add_error(self.all_error_key,"unable to save language.xml")
            raise
        self.cacheFile(languageXmlUrl,"languages.xml")
        
        
#---XML DOM Caching

    def getXmlDomUrl(self,url):
        cachedDom=self.domDict.get(url)
        if cachedDom:
            return cachedDom
        else:
            try:
                xmlDom=minidom.parse(getUrlRef(url))
            except urllib2.URLError:
                logging.error("unable to get XML reference from %s",url)
                raise

            self.domDict[url]=xmlDom
            return xmlDom

    def getXmlDomPath(self,path):
        cachedDom=self.domDict.get(path)
        if cachedDom:
            return cachedDom
        else:
            xmlDom=self.getXmlDomFile(open(path,"rb"))
            self.domDict[path]=xmlDom
            return xmlDom
        
    def getXmlDomFile(self,file):
        cachedDom=self.domDict.get(file)
        if cachedDom:
            return cachedDom
        else:
            xmlDom=minidom.parse(file)
            self.domDict[file]=xmlDom
            return xmlDom

    def getXmlDomString(self,string):
        cachedDom=self.domDict.get(string)
        if cachedDom:
            return cachedDom
        else:
            xmlDom=minidom.parseString(string)
            self.domDict[string]=xmlDom
            return xmlDom
        
    def getXmlDomZip(self,zip,zippedFile):
        cachedDom=self.domDict.get(zip)
        if cachedDom:
            return cachedDom
        else:
            xmlDom=self.getXmlDomString(zip.read(zippedFile))
            self.domDict[zip]=xmlDom
            return xmlDom


#------------------------------------------------
        
    def shapeUrl(self,base,afterKey):
        return urljoin(base,"/api/%s/%s"%(self.apiKey,afterKey))
    
    def randMirror(self,type):
        try:
            if "mirrors" not in self.__dict__:
                self.initMirrors()
        except:
            raise
        
        try:
            random_index=random.randint(0,len(self.mirrors[type])-1)
            return self.mirrors[type][random_index].mirrorpath
        except AttributeError:
            self.add_error(self.all_error_key,"unable to select mirror")
            raise
