#!/usr/local/bin/python
# Rob Dennis - Lurchington Q. Delacroix
# Keasby - TV renamer
# Project Homepage: http://code.google.com/p/keasby/

#---Python Modules
import re
import os
import logging
#---Keasby
from keasby_options import cl_opt,cl_arg
from tv_class import TvEpisode

def shape_template(template):
    return template

def get_regex_list(regex_tuple,regex_function=re.search):
    regexList = []
    for regex,string in regex_tuple:
        logging.debug("regex: %s,string %s, groups:",regex,string)
        if not string:
            logging.debug("empty string, skipping")
            continue
        
        try:
            regexList.append(regex_function(regex,string))
            logging.debug(regex_function(regex,string).groupdict())
        except AttributeError:
            logging.debug("no MatchObject")
            pass

    return regexList

def join_regex_list(manual_regex_list,builtin_regex_list):
    return get_regex_list(manual_regex_list + builtin_regex_list)

def cut_from_string(piece_to_cut,base_string):
    if piece_to_cut.isdigit():
        logging.debug("an all-digit piece specified")

        try:
            reshaped_cut = r"0*"+str(int(piece_to_cut))
            logging.debug("cutting '%s' from %s",reshaped_cut,base_string)
            new_string = re.sub(reshaped_cut,"",base_string,1)
            logging.debug("resulting in %s",new_string)
        
        except AttributeError:
            logging.debug(pieceToCut,"not found in",base_string)
            return basestring
        
        return new_string

        
    else:
        try:
            reshaped_cut = re.sub(r"\W+","\W+",piece_to_cut)
            logging.debug("cutting '%s' from %s",piece_to_cut,base_string)
            new_string = re.sub(r"(?i)"+reshaped_cut,"",base_string,1)
            logging.debug("resulting in %s",new_string)
        
        except AttributeError:
            logging.debug(pieceToCut,"not found in",base_string)
            return basestring
        
        return new_string
    
    
def get_path_parents(given_path):
    path,filename = os.path.split(given_path)
    path,parent_dir = os.path.split(path)
    path,parent_of_parent_dir = os.path.split(path)
    
    return parent_of_parent_dir,parent_dir,filename

    
    

class MediaInterpreter():
    def __init__(self,media_path,information_repository,parent,**manual_info):
        logging.debug("instantiated a media intepreter for %s",media_path)
        
        self.media_file = TvEpisode(information_repository,**manual_info)
        self.parent = parent
        
        path,filename = os.path.split(media_path)
        path,parent_dir = os.path.split(path)
        path,parent_of_parent_dir = os.path.split(path)
        
        pp_dir,p_dir,filename = get_path_parents(media_path)
        

        self.__update_info_key( "filename",filename,True)
        self.__update_info_key( "parent_dir",p_dir,True)                                            
        self.__update_info_key( "parent_of_parent_dir",pp_dir,True)
        
        self.__update_info_key( "path",
                                os.path.join(pp_dir,p_dir,filename),
                                force = True)
                                                
        self.__update_media_key("abspath",media_path,force = True)

        self.search_regex_dict = {}
        
        self.init_search_regex()
        
        self.errors = {}
        
    def init_search_regex(self):
        sep = {"sep": "\\"+os.sep}
        
        show_regex_tuple_list = [
            (r"(?i)(?P<showname>[^%(sep)s]+)%(sep)s\W*(?=(Season|Series))"%sep,
                                                            self["path"]),
            
            (r"(?i)(?P<showname>.+?)\W+(?:Season|Series)[^%(sep)s]*?\d+"%sep,
                                                        self["parent_dir"]),
            
            (r"(?i)(?P<showname>[^%(sep)s]+)(Season|Series|[^A-Z]s)[\w\s]*\d+"%sep,
                                                        self["path"]),
            
            (r"(?i)(?P<showname>.*)",self["parent_dir"])
            
            ]
                    
        self.search_regex_dict["showname"] = show_regex_tuple_list
        
        season_regex_tuple_list = []
                
        self.search_regex_dict["season_number"] = season_regex_tuple_list
        
        episode_regex_tuple_list = [
                (r"(?i)(?<=e])(?P<episode_number>[0-9]+)",self["filename"])
                ]
                
        self.search_regex_dict["episode_number"] = episode_regex_tuple_list
        

#---Parse Filename
    def get_media(self,template_list=None):
        if template_list:
            try:
                self.__update_media(
                                self.use_template(template_list).groupdict()
                                )
            except:
                logging.debug("no matching template")
                self.add_error("no matching template")
            
        else:            
            self.__auto_match(self.match_showname,
                                            "error guessing showname")
                                            
            self.__auto_match(self.match_season_number,
                                            "error guessing season_number")
                                            
            self.__auto_match(self.match_episode_number,
                                            "error guessing episode_number")
        
            
        return self.media_file
               
                
    def match_tagname(self,tagname,*search_expression,**kwargs):
        return join_regex_list( list(search_expression),
                                self.search_regex_dict[tagname])
                                
    def guess_tagname(self,tagname,match_function,*args,**kwargs):
        
        try:
            return match_function(*args,**kwargs).group(tagname)
        except AttributeError:
            return None
            

        

    def guess_showname(self):
        
        return self.guess_tagname("showname",self.match_showname)
        
    
    def guess_season_number(self):
        
        return self.guess_tagname("season_number",self.match_season_number)
                                
    def guess_episode_number(self):
        
        return self.guess_tagname("episode_number",self.match_episode_number)
        
    
    def match_showname(self):

        manual_showname = cl_opt.manual_showname
        
        if manual_showname is None:
            shownameMatches = self.match_tagname("showname")
            shownameMatch = self.__chose_match(shownameMatches)
        else:
            shownameMatch = re.search(r"(?P<showname>%s)"%manual_showname,
                                                          manual_showname)

        cut_filename = cut_from_string(shownameMatch.group("showname"),
                                                            self["filename"])

        self.__update_info_key("filename-showname",cut_filename,True)
        
        return shownameMatch
        
    def match_season_number(self):
        
        manual_season = cl_opt.manual_season
        if manual_season is None:
            season_number_matches = self.match_tagname("season_number")
           
            if cl_opt.assume_grouped_numbers:
                repeat_amount = r"+?((?=\d{%s})|(?=\D))"%cl_opt.grouped_num_split
            else:
                repeat_amount = r"+"
                
            cut_filename = self["filename-showname"]
            if not cut_filename:
                cut_filename = self["filename"]

            followup_season_regex_list = [
                    (r"(?i)(?=(Season|Series))\D*(?P<season_number>\d+)",
                                                               self["parent_dir"]), 
                    (r"(?i)(?<=[-\ _xs\.])[^\d()]?(?P<season_number>\d+)",
                                                               self["parent_dir"]),
                    (r"(?i)(?P<season_number>^\d%s)"%repeat_amount,cut_filename),

                    (r"(?i)(?<=(Season))\D*(?P<season_number>\d%s)"%repeat_amount,
                                                                     cut_filename),            
                    (r"(?i)(?<=[-\ _xs\.])(?P<season_number>\d%s)"%repeat_amount,
                                                                    cut_filename),
                    ]
            
            followup_season_matches = get_regex_list(followup_season_regex_list)
            season_match = self.__chose_match(followup_season_matches + 
                                                            season_number_matches)
            
            if cl_opt.handle_absolute_ordering:
                fullSeriesRegexList = [(r"(?i).*(?P<season_number>full).*",
                                     self["path"])
                                    ]
                fullSeriesMatches = get_regex_list(fullSeriesRegexList)
                fullSeriesMatch = self.__chose_match(fullSeriesMatches)
                
                if fullSeriesMatch:
                    ##it's a full run, using absolute ordering
                    return fullSeriesMatch
                
        else:
            re.search(r"(?P<manual_season>%s)"%manual_season,manual_season)

        remaining_numbers=re.findall(r"\d+", cut_filename)
        
        logging.debug("leftover numbers: %s",remaining_numbers)
        
        if (len(remaining_numbers) > 1 
            or len(remaining_numbers[0]) > cl_opt.grouped_num_split):
            
            cut_filename = cut_from_string(
                            season_match.group("season_number"),cut_filename
                                )
        
            self.__update_info_key("filename-showname-season",
                                cut_filename,
                                force = True)
                                
        else:
            logging.debug("assuming that remaining number shouldn't be cut")
        
        return season_match
        
    def match_episode_number(self,*arg,**kwargs):
  
        ep_matches = self.match_tagname("episode_number",*arg,**kwargs)
        
        cut_filename = self["filename-showname-season"]
        
        if not cut_filename:
            cut_filename = self["filename-showname"]
        if not cut_filename:
            cut_filename = self["filename"]
        
        followup_episode_regex_list = [
        
                (r"(?i)(?<=e])(?P<episode_number>\d+(?:\%s+\d+)*)"%
                                cl_opt.split_episode_char,self["filename"]),
                                
                (r"(?i)(?<=[\ _xe\.])(?P<episode_number>\d+(?:\%s+\d+)*)"%
                                cl_opt.split_episode_char,cut_filename),
                                
                (r"(?i)(?P<episode_number>\d+(?:\%s+\d+)*)"%
                                cl_opt.split_episode_char,cut_filename),
                                
                ]
        
        followup_episode_matches = get_regex_list(followup_episode_regex_list)
        
        episode_number_match = self.__chose_match(followup_episode_matches + 
                                                        ep_matches)
        
        
        
        return episode_number_match
        
    def use_template(self,template_list):
        regex_template_list = [
                (r"(?P<filename>.*)\.",self["filename"])
                    ]
        regex_template_list = [(shape_template(template),self["path"]) 
                                    for template in template_list]
        template_matches = join_regex_list(regex_template_list,regex_template_list)
        
        return self.__chose_match(template_matches)
    
    def search_title(self):
        logging.debug("searching title from MediaInterpreter")
        self.get_media()
            
        title_list = self.media_file.get_titles()
        
        for episode,title in self.media_file.episode_dict.items():
            logging.debug("checking title %s for episode %s"%(title,episode))
            if title is None:            
                self.add_error("unable to search title using: %s s:%s e:%s"%
                                (self.media_file["showname"],
                                self.media_file["season_number"],
                                episode
                                )
                            )
            
        return self.media_file
        
        
        
#---Diagnostic

    def has_errors(self):
        return len(self.errors.keys())>0
    
    def add_error(self,error_string,message=None):
        if not message:
            message = self["path"]

        if error_string in self.errors:
            self.errors[error_string].append(message)
        else:
            self.errors[error_string] = [message]
            
#---Utility    
    def __auto_match(self,guess_function,error_message=None):
        try:
            returned_dict = guess_function().groupdict()
            logging.debug(returned_dict)
            self.__update_media(returned_dict)
        except:
            if not error_message:
                error_message = "error in %s"%guess_function.__name__
            logging.debug(error_message)
            self.add_error(error_message)

    def __user_choice(self,options):
        for option in options:
            if option:
                return option

    def __chose_match(self,matches):
        choice = self.__user_choice(matches)

        return choice
    
    def __update_media_key(self,key,value,force=False):
        if force:
            self.media_file[key] = value
        else:
            logging.debug("setting %s to |%s| in interpreter",key,value)
            self.media_file[key] = self.__user_choice([value,self.media_file[key]])
            
    def __update_media(self,information_dict,force=False):
        for key,value in information_dict.items():
            self.__update_media_key(key,value,force)

    def __update_info_key(self,key,value,force=False):
        if force:
            self[key] = value
        else:
            self[key] = self.__user_choice([value,self[key]])
    
    def __update_info(self,information_dict,force=False):
        for key,value in information_dict.items():
            self.__update_info_key(key,value,force)
        
    def __getitem__(self,key):
        return self.__dict__.get(key)
    
    def __setitem__(self,key,value):
        return self.__dict__.update({key:value})
        
    def __str__(self):
        return str(self.media_file)
        
